/*
 * HTTP_Server.c
 *
 * Created: 4/9/2022 6:20:15 PM
 *  Author: Zedan
 */ 

#define  F_CPU 8000000UL
#include <util/delay.h>
#include <stdint.h>
#include <avr/pgmspace.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "Z_USART.h"
#include "Z_SPI.h"
#include "w5500.h"
#include "socket.h"
#include "webpage.h"
#include "HTTP_Server.h"


#define HTTPD_BUF_SIZE	128
#define MAX_URI_SIZE 15
#define MAX_URI_SIZE 15


/*********************************************
* HTTP Method 
*********************************************/
#define		METHOD_ERR		0		/**< Error Method. */
#define		METHOD_GET		1		/**< GET Method.   */
#define		METHOD_HEAD		2		/**< HEAD Method.  */
#define		METHOD_POST		3		/**< POST Method.  */


//uint8_t http_request[HTTPD_BUF_SIZE];												
//uint8_t	http_ParsedRequest_METHOD;
//char http_ParsedRequest_URI[MAX_URI_SIZE];

struct _st_http_request
{
	uint8_t http_request[HTTPD_BUF_SIZE];
	uint8_t	http_ParsedRequest_METHOD;					/**< request method(METHOD_GET...). */
	char http_ParsedRequest_URI[MAX_URI_SIZE];			/**< request file name.             */
};

struct _st_http_request HttpInfo[MAX_HTTPSOCK];

void httpServer_run(uint8_t seqnum);
//static void parse_http_request(uint8_t * request);
void parse_http_request(uint8_t s, uint8_t * request);
static void http_handler(uint8_t s, char* p_http_request);
static void html_send(uint8_t s, char* requested_uri);
int32_t NetSend(uint8_t sn, uint8_t * buf, uint16_t len);

void httpServer_run(uint8_t seqnum)
{
	uint8_t s;	// socket number
	uint16_t len;

	// Get the H/W socket number
	s = seqnum;  

	/* HTTP Service Start */
	switch(getSn_SR(s))
	{
		case SOCK_ESTABLISHED:
			// Interrupt clear
			if(getSn_IR(s) & Sn_IR_CON)	{setSn_IR(s, Sn_IR_CON);}		
			// HTTP Process states			
			if ((len = getSn_RX_RSR(s)) > 0)
				{
				if (len > HTTPD_BUF_SIZE) len = HTTPD_BUF_SIZE;
				len = recv(s, HttpInfo[s].http_request, len);
				
				*((HttpInfo[s].http_request) + len) = '\0';  // insert null-terminate symbol to correct parse data
				
				parse_http_request(s, HttpInfo[s].http_request);
				http_handler(s, HttpInfo[s].http_ParsedRequest_URI);
				char StrBuf[100];
				sprintf(StrBuf,"Request received on socket %d\n\r",s);
				UART_SendString(StrBuf);
				// Check the TX socket buffer for End of HTTP response sends
				while(getSn_TX_FSR(s) != (getSn_TXBUF_SIZE(s)*1024));
				}
	
			disconnect(s);

		break;

		case SOCK_CLOSE_WAIT:
			disconnect(s);
		break;

		case SOCK_CLOSED:
			if(socket(s, Sn_MR_TCP, 80, 0x00) == s)    /* Recreate the socket the socket */
				{ /* SOCKET recreate successfully */ }
		break;

		case SOCK_INIT:
			listen(s);
		break;

		case SOCK_LISTEN:
		break;

		default :
		break;

	} // end of switch

}


void parse_http_request(uint8_t s, uint8_t * request)
{
	char * nexttok;
	nexttok = strtok((char*)request," ");
	
	if(!nexttok) { HttpInfo[s].http_ParsedRequest_METHOD = METHOD_ERR;  return;	}
		
	if      (!strcasecmp(nexttok, "GET"))  { HttpInfo[s].http_ParsedRequest_METHOD = METHOD_GET;  nexttok = strtok(NULL," "); }
	else if (!strcasecmp(nexttok, "HEAD")) { HttpInfo[s].http_ParsedRequest_METHOD = METHOD_HEAD; nexttok = strtok(NULL," "); }
	else if (!strcasecmp(nexttok, "POST")) { nexttok = strtok(NULL,"\0"); HttpInfo[s].http_ParsedRequest_METHOD = METHOD_POST; }
	else { HttpInfo[s].http_ParsedRequest_METHOD = METHOD_ERR;	}
	
	strcpy((char *)HttpInfo[s].http_ParsedRequest_URI, nexttok);
}


static void http_handler(uint8_t s, char* p_http_request_uri)
{
	//method Analysis
	switch (HttpInfo[s].http_ParsedRequest_METHOD)
	{
		case METHOD_HEAD :
		case METHOD_GET :	
			html_send(s, p_http_request_uri);		
		break;

		case METHOD_POST :	
		break;

		default :
		break;
	}
}

void html_send(uint8_t s, char* requested_uri)
{
	if (!strcmp(requested_uri, "/")) NetSend(s, (uint8_t *)HOME_PAGE, strlen((char*)HOME_PAGE));
	if (!strcmp(requested_uri, "/favicon.ico")) NetSend(s, (uint8_t *)FAVICON, strlen((char*)FAVICON));  
	//else UART_SendString("Page Not Found \n\r");
}

	
int32_t NetSend(uint8_t sn, uint8_t * buf, uint16_t len)
{
	uint16_t freesize=0;
	freesize = getSn_TxMAX(sn);
	while(1)
		{
		freesize = getSn_TX_FSR(sn);
		if(len <= freesize) break;
		}
	
	uint16_t ptr = 0;
	uint32_t addrsel = 0;

	//if(len == 0)  return;
	ptr = getSn_TX_WR(sn);
	addrsel = ((uint32_t)ptr << 8) + (WIZCHIP_TXBUF_BLOCK(sn) << 3);
	
	uint16_t i;

	//WIZCHIP_CRITICAL_ENTER();
	W5500_Select();

	addrsel |= (_W5500_SPI_WRITE_ | _W5500_SPI_VDM_OP_); //Combining Address & Control
	
	//Sending Address & Control
	SPI_SndByte((addrsel & 0x00FF0000) >> 16);  //Address High Byte
	SPI_SndByte((addrsel & 0x0000FF00) >>  8);  //Address low Byte
	SPI_SndByte((addrsel & 0x000000FF) >>  0);	//Control Phase
		
	for(i = 0; i < len; i++)    SPI_SndByte(pgm_read_byte(&buf[i]));   //SPI_SndByte(buf[i]);
		
	W5500_DeSelect();
	//WIZCHIP_CRITICAL_EXIT();
	
	ptr += len;
	setSn_TX_WR(sn,ptr);

	/* -------------------------- Test Code -------------------------------------*/
	/*
	ptr = 0;
	addrsel = 0;
	
	ptr = getSn_TX_RD(sn);
	addrsel = ((uint32_t)ptr << 8) + (WIZCHIP_TXBUF_BLOCK(sn) << 3);
	
	W5500_Select();

	addrsel |= (_W5500_SPI_READ_  | _W5500_SPI_VDM_OP_);
	
	uint8_t * tempbuf ={'\0', };
		
	//Sending Address & Control
	SPI_SndByte((addrsel & 0x00FF0000) >> 16);  //Address High Byte
	SPI_SndByte((addrsel & 0x0000FF00) >>  8);  //Address low Byte
	SPI_SndByte((addrsel & 0x000000FF) >>  0);	//Control Phase
	
	for(i = 0; i < len; i++)    tempbuf[i]=SPI_RcvByte();
	tempbuf[i+1]='\0';
	W5500_DeSelect();
	//for(i = 0; i < len; i++)    UART_SendChar(pgm_read_byte(&buf[i]));
	UART_SendString("\n\r##################################################\n\r"); 
	UART_SendString((char*)tempbuf); UART_SendString("\n\r");
	if (!(strcmp((char*)tempbuf,(char*)buf))) UART_SendString("Equal\n\r");  else UART_SendString("NOT Equal\n\r");
	*/
	/* ----------------------- End of Test Code ---------------------------------*/

	setSn_CR(sn,Sn_CR_SEND);  //Sending the data in the buffer
	
	/* wait to process the command... */
	while(getSn_CR(sn));
	
	return (int32_t)len;
}

